; Copyright (c) 2020 Alan Dipert <alan@dipert.org>
; Part of the JACL project: https://tailrecursion.com/JACL/
;
; Permission is hereby granted, free of charge, to any person obtaining a copy
; of this software and associated documentation files (the "Software"), to deal
; in the Software without restriction, including without limitation the rights
; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
; copies of the Software, and to permit persons to whom the Software is
; furnished to do so, subject to the following conditions:
;
; The above copyright notice and this permission notice shall be included in all
; copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
; OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
; SOFTWARE.

(jacl:enable-js-syntax)

(let* ((jpkg (jacl:\. (jacl:%js "Package") (|get| @"JACL"))))
  (jacl:\. *package* (|usePackage| jpkg)))

(defvar *is*)

(defmacro deftest (label &rest body)
  (let ((is (gensym)))
    `(\. @|QUnit| (|test| (\. ,label (|toString|))
                   (lambda (,is)
                     (declare (jacl:async))
                     (let ((*is* ,is))
                       ,@body))))))

(defun assert= (x y &optional (label ""))
  (\. *is* (|strictEqual| x y (\. label (|toString|)))))

(defmacro assert-throws (&rest body)
  `(\. *is* (|throws| (lambda () ,@body))))

(defun read1 (s)
  (let ((s (\. s (|toString|))))
    (\. (%new @|Reader| (%new @|StringStream| s)) (|read|))))

(defun log (&rest objects)
  (declare (jacl:rest-array))
  (\. (jacl:%js "console") |log| (|apply| nil objects)))

(\. @|QUnit| (|module| @"Reader"))

(deftest "Integers"
  (assert= (await (read1 "123 "))   123  "single integer")
  (assert= (await (read1 "+9912 ")) 9912 "integer with leading +")
  (assert= (await (read1 "0 "))     0    "zero")
  (assert= (await (read1 "-32 "))   -32  "negative number")
  (assert= (await (read1 "1. "))    1.   "number with trailing dot"))

(deftest "Basic Symbols"
  (let ((sym (await (read1 @"somesym "))))
    (assert= (\. sym |name|) @"SOMESYM" "simple symbol name")
    (assert= (\. sym |packageName|) @"COMMON-LISP-USER" "simple symbol package"))
  (let ((sym (await (read1 @"|Alan| "))))
    (assert= (\. sym |name|) @"Alan" "simple symbol name"))
  ;; TODO Support escape syntax in the JavaScript String reader macro.
  (let ((sym (await (read1 @"\\\\alan "))))
    (assert= (\. sym |name|) @"aLAN" "simple symbol with escape")))

(\. @|QUnit| (|module| @"Numerics"))

(deftest "Numeric functions"
  (assert= (zerop 1) nil)
  (assert= (zerop 0) t)

  (assert= (1+ 2) 3)
  (assert= (1- 10) 9)

  (assert-throws (>))
  (assert-throws (> "foo"))
  (assert-throws (> "foo" "bar"))
  (assert-throws (> "foo" "bar" "baz"))
  (assert= (> 3 2 1) t)
  (assert= (> 1) t)
  (assert= (> 1 3) nil)

  (assert-throws (<))
  (assert-throws (< "foo"))
  (assert-throws (< "foo" "bar"))
  (assert-throws (< "foo" "bar" "baz"))
  (assert= (< 1) t)
  (assert= (< 5 10 421) t)
  (assert= (< 41 2) nil))

(\. @|QUnit| (|module| @"Control operators"))

(deftest "Local TAGBODY"
  (let ((x 10)
        (y 0))
    (tagbody
     begin
     (when (> x 0)
       (setq x (1- x) y (1+ y))
       (go begin))
     end)
    (assert= x 0)
    (assert= y 10)))

(deftest "GO in TAGBODY prelude"
  (let ((x 0))
    (tagbody
     (go end)
     start                              ;skipped
     (setq x (1+ x))                    ;skipped
     end)
    (assert= x 0)))

(defun countdown (from)
  (tagbody
   start
   (when (> from 0)
     (setq from (1- from))
     (go start))))

(countdown 3)

(\. @|QUnit| (|start|))
